from typing import Dict, Union
import inspect
import os
import sys
import pickle
import tempfile

from ..lib.gravity import Module
from ..lib.ci_exception import SilentAbortException

__all__ = [
    "ApiSupport"
]


class ApiSupport(Module):
    def __init__(self, api_mode: bool = False, **kwargs) -> None:
        super().__init__(**kwargs)

        if api_mode:
            if "UNIVERSUM_DATA_FILE" not in os.environ:
                raise EnvironmentError(
                    inspect.cleandoc("""Error: Failed to read the 'UNIVERSUM_DATA_FILE' from environment
                
                    This command is intended to be run from within Universum run (e.g. as a separate step
                    in project config file). If you got this message by running it from command line:
                    please don't. If you got this message by running it with Universum: something must
                    have gone wrong, may be a bug in Universum itself. Feel free to contact the developers."""))

            with open(os.getenv("UNIVERSUM_DATA_FILE", ""), "rb+") as data_file:
                self.data = pickle.load(data_file)
        else:
            self.data_file = tempfile.NamedTemporaryFile(mode="wb+")  # pylint: disable = consider-using-with
            self.data = {}

    def _set_entry(self, name: str, entry: Union[str, bool]) -> None:
        self.data[name] = entry

    def _get_entry(self, name: str) -> str:
        return self.data.get(name, "")

    def get_environment_settings(self) -> Dict[str, str]:
        pickle.dump(self.data, self.data_file)
        self.data_file.flush()
        return {"UNIVERSUM_DATA_FILE": self.data_file.name}

    def add_file_diff(self, entry: Union[str, None]) -> None:
        if entry is None:
            self._set_entry("DIFF_FAILED", True)
        else:
            self._set_entry("DIFF", entry)

    def get_file_diff(self) -> str:
        if self._get_entry("DIFF_FAILED") is True:
            sys.stderr.write("Getting file diff failed due to Perforce server internal error")
            raise SilentAbortException(application_exit_code=1)
        return self._get_entry("DIFF")



import datetime
import importlib
import urllib.parse

from . import git_vcs
from ..error_state import HasErrorState
from ..reporter import ReportObserver, Reporter
from ...lib import utils
from ...lib.gravity import Dependency

__all__ = [
    "GithubToken",
    "GithubMainVcs"
]

github = None


def get_time():
    return datetime.datetime.utcnow().replace(microsecond=0).isoformat() + "Z"


class GithubToken(HasErrorState):

    @staticmethod
    def define_arguments(argument_parser):
        parser = argument_parser.get_or_create_group("GitHub", "GitHub repository settings")

        parser.add_argument("--github-app-id", "-gta", dest="integration_id", metavar="GITHUB_APP_ID",
                            help="GitHub application ID to use for check run report. Only GitHub App "
                                 "can report a check run result! If you don't have an App for reporting purposes, "
                                 "please don't use ``--report-to-review`` with GitHub")
        parser.add_argument("--github-private-key", "-gtk", dest="key", metavar="GITHUB_PRIVATE_KEY",
                            help="GitHub App private key for obtaining installation authentication token. "
                                 "Pass raw key data via environment variable or pass a file path to read the key from "
                                 "by starting the value string with '@'. File path can be either absolute or relative. "
                                 "Please note, that when passing a file, it is expected to be in UTF-8 encoding")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.check_required_option("integration_id", """
            The GitHub App ID is not specified.

            Only GitHub Application owner knows this ID. If you are the App owner, please
            check your App's general settings. If not, please contact the App owner for this
            information.

            Please note that 'universum github-handler' DOES NOT pass App ID to CI builds.
            It is assumed that one CI configuration (job, build, workflow) serves as one
            GitHub App. Because of that, it is required to specify App ID within the CI
            configuration.

            Please specify the GitHub App ID by using '--github-app-id' ('-gta') command
            line parameter or by setting GITHUB_APP_ID environment variable.
            """)

        self.key = self.read_and_check_multiline_option("key", """
            The GitHub App private key is not specified.

            Please note that 'universum github-handler' DOES NOT pass private key to CI
            builds. It is assumed that one CI configuration (job, build, workflow) serves as
            one GitHub App. Because of that, it is required to specify private key within
            the CI configuration.

            As the private key is a multiline string, it is not convenient to pass it
            directly via command line. If you start the parameter with '@', the rest should be
            the path to a file containing the key. The path can be absolute or relative to the 
            project root. You can also store the key in GITHUB_PRIVATE_KEY environment variable.
            Please note, that when passing a file, it's expected to be in UTF-8 encoding")
            """)

        global github
        try:
            github = importlib.import_module("github")
        except ImportError as e:
            text = "Error: using GitHub Handler or VCS type 'github' requires Python package 'pygithub' " \
                   "to be installed to the system for correct GitHub App token processing. " \
                   "It also requires Python package 'cryptography' to be installed in addition. " \
                   "Please refer to 'Prerequisites' chapter of project documentation for detailed instructions"
            raise ImportError(text) from e

        self.token_issued = None
        self._token = None

    def _get_token(self, installation_id):
        integration = github.GithubIntegration(self.settings.integration_id, self.key)
        auth_obj = integration.get_access_token(installation_id)
        return auth_obj.token

    def get_token(self, installation_id):
        if self._token:
            token_age = (datetime.datetime.now() - self.token_issued).total_seconds() / 60
            if token_age < 55:  # GitHub token lasts for one hour
                return self._token
        self._token = self._get_token(installation_id)
        self.token_issued = datetime.datetime.now()
        return self._token


class GithubTokenWithInstallation(GithubToken):

    @staticmethod
    def define_arguments(argument_parser):
        parser = argument_parser.get_or_create_group("GitHub", "GitHub repository settings")

        parser.add_argument("--github-installation-id", "-gti", dest="installation_id",
                            metavar="GITHUB_INSTALLATION_ID",
                            help="GitHub installation ID identifies specific app installation into user account "
                                 "or organization. Can be retrieved from web-hook or obtained via REST API; "
                                 "in standard workflow should be received from 'universum github-handler'")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.check_required_option("installation_id", """
            The GitHub App installation ID not specified.

            An installation refers to any user or organization account that has installed
            the app. Even if someone installs the app on more than one repository, it only
            counts as one installation because it's within the same account. Installation ID
            can be retrieved via REST API or simply parsed from GitHub App web-hook.

            If using 'universum github-handler', installation ID is automatically extracted
            from the webhook payload and  passed via GITHUB_INSTALLATION_ID environment
            variable.
            """)

    def get_token(self, installation_id=None):
        return super().get_token(installation_id=self.settings.installation_id)


class GithubMainVcs(ReportObserver, git_vcs.GitMainVcs, GithubTokenWithInstallation):
    """
    This class mostly contains functions for Gihub report observer
    """
    reporter_factory = Dependency(Reporter)

    @staticmethod
    def define_arguments(argument_parser):
        parser = argument_parser.get_or_create_group("GitHub", "GitHub repository settings")

        parser.add_argument("--github-check-name", "-ghc", dest="check_name", metavar="GITHUB_CHECK_NAME",
                            default="Universum check", help="The name of Github check run")
        parser.add_argument("--github-check-id", "-ghi", dest="check_id", metavar="GITHUB_CHECK_ID",
                            help="Github check run ID")
        parser.add_argument("--github-api-url", "-gha", dest="api_url", metavar="GITHUB_API_URL",
                            default="https://api.github.com/", help="API URL for integration")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reporter = None

        self.check_required_option("checkout_id", """
            The git checkout id for github is not specified.

            For github the git checkout id defines the commit to be checked and reported.
            Please specify the checkout id by using '--git-checkout-id' ('-gco') command
            line parameter or by setting GIT_CHECKOUT_ID environment variable.

            If using 'universum github-handler', the checkout ID is automatically extracted
            from the webhook payload and passed via GIT_CHECKOUT_ID environment variable.
            """)

        self.check_required_option("check_id", """
            The GitHub Check Run ID is not specified.

            GitHub check runs each have unique ID, that is used to update check result. To
            integrate Universum with GitHub, check run should be already created before
            performing actual check. Please specify check run ID by using
            '--github-check-id' ('-ghi') command line parameter or by setting
            GITHUB_CHECK_ID environment variable.

            If using 'universum github-handler', the check ID is automatically extracted
            from the webhook payload and passed via GITHUB_CHECK_ID environment variable.
            """)

        self.request = {}
        self.request["status"] = "in_progress"
        self.request["output"] = {
            "title": self.settings.check_name,
            "summary": ""
        }

    def _clone(self, history_depth, destination_directory, clone_url):
        parsed_repo = urllib.parse.urlsplit(clone_url)
        if parsed_repo.scheme == "https" and not parsed_repo.username:
            new_netloc = f"x-access-token:{self.get_token()}@{parsed_repo.netloc}"
            parsed_repo = (parsed_repo.scheme, new_netloc, parsed_repo.path, parsed_repo.query, parsed_repo.fragment)
        clone_url = urllib.parse.urlunsplit(parsed_repo)
        super()._clone(history_depth, destination_directory, clone_url)

    def code_review(self):
        self.reporter = self.reporter_factory()
        self.reporter.subscribe(self)
        return self

    def update_review_version(self):
        self.out.log("GitHub has no review versions")

    def get_review_link(self):
        return self.settings.repo.rsplit(".git", 1)[0] + "/runs/" + self.settings.check_id

    def is_latest_version(self):
        return True

    def _report(self):
        repo_path = str(urllib.parse.urlsplit(self.settings.repo).path).rsplit(".git", 1)[0]
        check_url = self.settings.api_url + "repos" + repo_path + "/check-runs/" + self.settings.check_id

        headers = {
            "Accept": "application/vnd.github.antiope-preview+json",
            "Authorization": "token " + self.get_token()
        }

        utils.make_request(check_url, request_method="PATCH", json=self.request, headers=headers, timeout=5*60)

    def code_report_to_review(self, report):
        # git show returns string, each file separated by \n,
        # first line consists of commit id and commit comment, so it's skipped
        commit_files = self.repo.git.show("--name-only", "--oneline", self.settings.checkout_id).split('\n')[1:]
        comments = []
        for path, issues in report.items():
            if path in commit_files:
                for issue in issues:
                    comments.append(dict(path=path,
                                         message=issue['message'],
                                         start_line=issue['line'],
                                         end_line=issue['line'],
                                         annotation_level="warning"))
        self.request["output"]["annotations"] = comments
        self._report()

    def report_start(self, report_text):
        self.request["started_at"] = get_time()
        self.request["output"]["summary"] = report_text
        self._report()

    def report_result(self, result, report_text=None, no_vote=False):
        if result:
            conclusion = "success"
        else:
            conclusion = "failure"

        if not report_text:
            report_text = ""

        self.request["status"] = "completed"
        self.request["completed_at"] = get_time()
        self.request["conclusion"] = conclusion
        self.request["output"]["summary"] = report_text
        self._report()



from typing import List

from .terminal_based_output import TerminalBasedOutput

__all__ = [
    "GithubOutput"
]


class GithubOutput(TerminalBasedOutput):
    """
    GitHub Actions manual: https://docs.github.com/en/actions
    GitHub doesn't support nested grouping (https://github.com/actions/runner/issues/802)
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._block_opened: bool = False

    def _print_lines(self, *args, **kwargs) -> None:
        prefix: str = kwargs.setdefault("prefix", "")
        result: str = "".join(args)
        lines: List[str] = result.splitlines(False)
        for line in lines:
            self._stdout(f"{prefix}{line}")

    def log_error(self, description: str) -> None:
        self._print_lines(description, prefix="::error::")

    def log_stderr(self, line: str) -> None:
        self._print_lines("stderr: ", line, prefix="::warning::")

    def open_block(self, num_str: str, name: str) -> None:
        if self._block_opened:
            self._print_lines("::endgroup::")

        self._print_lines(f"::group::{num_str} {name}")
        self._block_opened = True

    def close_block(self, num_str: str, name: str, status: str) -> None:
        if self._block_opened:
            self._block_opened = False
            self._print_lines("::endgroup::")

        if status == "Failed":
            self._print_lines(f"::error::{num_str} {name} - Failed")

    def log_skipped(self, message: str) -> None:
        self._print_lines(message, prefix="::warning::")



import glob
import json
import os
import urllib.parse
from copy import deepcopy
from typing import Dict, List, Optional, TextIO, Tuple

from . import artifact_collector, reporter
from .output import HasOutput
from .project_directory import ProjectDirectory
from .structure_handler import HasStructure
from ..configuration_support import Configuration
from ..lib import utils
from ..lib.gravity import Dependency
from ..lib.utils import make_block


class CodeReportCollector(ProjectDirectory, HasOutput, HasStructure):
    reporter_factory = Dependency(reporter.Reporter)
    artifacts_factory = Dependency(artifact_collector.ArtifactCollector)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.artifacts: artifact_collector.ArtifactCollector = self.artifacts_factory()
        self.reporter: reporter.Reporter = self.reporter_factory()
        self.report_path: str = ""
        self.repo_diff: Optional[List[Tuple[Optional[str], Optional[str], Optional[str]]]]

    def prepare_environment(self, project_config: Configuration) -> Configuration:
        afterall_steps: Configuration = Configuration()
        for item in project_config.configs:
            if item.children:
                afterall_steps += self.prepare_environment(item.children)
            if not item.code_report:
                continue
            if not self.report_path:
                self.report_path = os.path.join(self.settings.project_root, "code_report_results")
                if not os.path.exists(self.report_path):
                    os.makedirs(self.report_path)
            temp_filename: str = "${CODE_REPORT_FILE}"
            name: str = utils.calculate_file_absolute_path(self.report_path, item.name) + ".json"
            actual_filename: str = os.path.join(self.report_path, name)

            item.replace_string(temp_filename, actual_filename)
            afterall_steps += [deepcopy(item)]
        return afterall_steps

    def _report_as_pylint_json(self, report) -> int:
        for result in report:
            text = result["symbol"] + ": " + result["message"]
            path = result["path"]
            message: reporter.ReportMessage = {"message": text, "line": int(result["line"])}
            self.reporter.code_report(path, message)
        return len(report)

    def _report_as_sarif_json(self, report) -> int:
        version: str = report.get('version', '')
        if version != '2.1.0':
            raise ValueError(f"Version {version} is not supported")
        issue_count: int = 0
        for run in report.get('runs', []):
            analyzer_data: Dict[str, str] = run.get('tool').get('driver')  # non-optional per definition
            who: str = f"{analyzer_data.get('name')} [{analyzer_data.get('version', '?')}]"
            for issue in run.get('results', []):
                issue_count += 1
                what: str = issue.get('message')
                for location in issue.get('locations', []):
                    location_data: Dict[str, Dict[str, str]] = location.get('physicalLocation')
                    if not location_data:
                        continue
                    artifact_data = location_data.get('artifactLocation')
                    if not artifact_data:
                        if location_data.get('address'):
                            continue  # binary artifact can't be processed
                        raise ValueError("Unexpected lack of artifactLocation tag")
                    path: str = urllib.parse.unquote(artifact_data.get('uri', ''))
                    region_data = location_data.get('region')
                    if not region_data:
                        continue  # TODO: cover this case as comment to the file as a whole
                    message: reporter.ReportMessage = {"message": f"{who} : {what}",
                                                       "line": int(region_data.get('startLine', '0'))}
                    self.reporter.code_report(path, message)
        return issue_count

    @make_block("Processing code report results")
    def report_code_report_results(self) -> None:
        reports: List[str] = glob.glob(self.report_path + "/*.json")
        for report_file in reports:
            with open(report_file, "r", encoding="utf-8") as f:
                text: str = f.read()
                report: Optional[List[Dict[str, str]]] = None
                if text:
                    report = json.loads(text)

            json_file: TextIO = self.artifacts.create_text_file("Static_analysis_report.json")
            json_file.write(json.dumps(report, indent=4))

            issue_count: int
            if not report and report != []:
                self.out.log_error("There are no results in code report file. Something went wrong.")
                continue

            try:
                issue_count = self._report_as_sarif_json(report)
            except (KeyError, AttributeError, ValueError):
                try:
                    issue_count = self._report_as_pylint_json(report)
                except (KeyError, AttributeError, ValueError):
                    self.out.log_error("Could not parse report file. Something went wrong.")
                    continue

            if issue_count != 0:
                text = str(issue_count) + " issues"
                self.out.log_error("Found " + text)
                self.out.set_build_status(os.path.splitext(os.path.basename(report_file))[0] + ": " + text)
            else:
                self.out.log("Issues not found.")



import shutil

from ..error_state import HasErrorState
from ..project_directory import ProjectDirectory
from ...lib.ci_exception import CiException
from ...lib.utils import make_block

__all__ = [
    "BaseVcs",
    "BaseDownloadVcs",
    "BaseSubmitVcs",
    "BasePollVcs"
]


class BaseVcs(ProjectDirectory):
    """
    Base class for VCS drivers
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.repo_status = ""
        self.sources_need_cleaning = False

    def append_repo_status(self, line):
        self.repo_status += line

    def get_repo_status(self):
        return self.repo_status

    @make_block("Cleaning copied sources", pass_errors=False)
    def clean_sources(self):
        try:
            shutil.rmtree(self.settings.project_root)
        except OSError as e:
            text = f"{e}\n"
            text += "\nPossible reasons of this error:" + \
                    "\n * Sources were not copied due to runtime errors" + \
                    "\n * Copied sources are already deleted while executing generated scenario" + \
                    "\n * Source vcs permissions do not allow deleting"
            raise CiException(text) from e

    def finalize(self):
        if self.sources_need_cleaning:
            self.clean_sources()


class BaseDownloadVcs(BaseVcs, HasErrorState):
    """
    Base class for default CI build VCS drivers
    """

    def code_review(self):
        return None

    def login(self):
        pass

    def prepare_repository(self):
        raise NotImplementedError

    def copy_cl_files_and_revert(self):
        raise NotImplementedError

    def calculate_file_diff(self):
        raise NotImplementedError


class BaseSubmitVcs(BaseVcs):
    """
    Base class for submitter VCS drivers
    """

    def submit_new_change(self, description, file_list, review=False, edit_only=False):
        """
        Create new change, add all reconciled vcs to index and submit to repository
        :param description: Change description
        :param file_list: List of full paths to vcs to be submitted
        :param review: When set to True, create deletable review instead of submit
        :param edit_only: When set to True, do not add new files to repository, just edit existing ones
        :return: Created change number
        """
        raise NotImplementedError


class BasePollVcs(BaseVcs):
    """
    Base class for poller VCS drivers
    """

    def get_changes(self, changes_reference=None, max_number='1'):
        """
        Get all (or last 'max_number' for a depot path) changes starting from reference state
        :param changes_reference: Dictionary, where keys are depot paths and values are the latest known changes
        :param max_number: Maximum number of changes for every depot path to return
        :return: Dictionary, where keys are depot paths and values are sets of changes
        """
        raise NotImplementedError



from ..error_state import HasErrorState
from ...lib.module_arguments import ModuleArgumentParser
from .base_server import BaseServerForHostingBuild, BaseServerForTrigger

__all__ = [
    "GithubServer",
]


class GithubServer(BaseServerForHostingBuild, BaseServerForTrigger, HasErrorState):
    @staticmethod
    def define_arguments(argument_parser: ModuleArgumentParser) -> None:
        parser = argument_parser.get_or_create_group(
            "GitHub Actions variables",
            "GitHub sets these environment variables and usually do not need to set these arguments. See: https://"
            "docs.github.com/en/actions/learn-github-actions/environment-variables#default-environment-variables"
        )
        parser.add_argument("--github-server", dest="server_url", metavar="GITHUB_SERVER_URL",
                            help="GitHub server URL, used on GitHub Actions for constructing links to builds. "
                                 "For example: https://github.com", type=str)
        parser.add_argument("--github-repo", dest="repo", metavar="GITHUB_REPOSITORY",
                            help="GitHub owner and repository name, used on GitHub Actions for constructing links to "
                                 "builds. For example: octocat/Hello-World", type=str)
        parser.add_argument("--github-run-id", dest="run_id", metavar="GITHUB_RUN_ID",
                            help="GitHub run ID, used on GitHub Actions for constructing links to builds."
                                 "For example: 1658821493", type=str)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.check_required_option("server_url", """
            The URL of the GitHub server is not specified.
            
            The URL of the GitHub server is used for constructing links to builds in code review system.

            Please specify the GitHub server url by using '--github-server'command-line  option or GITHUB_SERVER_URL 
            environment variable.
            """)
        self.check_required_option("repo", """
            The GitHub repository is not specified.
            
            The name of the GitHub repository is used for constructing links to builds in code review system.

            Please specify the GitHub repository by using '--github-repo' command-line option or 
            GITHUB_REPOSITORY environment variable.
            """)
        self.check_required_option("run_id", """
            The GitHub run ID is not specified.
            
            The id of the run on GitHub is used for constructing links to builds in code review system.

            Please specify the Github repository by using '--github-run-id' command-line option or 
            GITHUB_RUN_ID environment variable.
            """)

    def report_build_location(self) -> str:
        build_link: str = f"{self.settings.server_url}/{self.settings.repo}/actions/runs/{self.settings.run_id}"
        return "Here is the link to Github Actions build: " + build_link

    def artifact_path(self, local_artifacts_dir: str, item: str) -> str:
        return item
